use chrono::{Days, NaiveDateTime};
use sqlx::PgPool;

use super::Availability;
use crate::{validation::{
    start_date_time_lies_before_end_date_time, AsyncValidate, AsyncValidateError
}, END_OF_DAY, START_OF_DAY};

pub struct AvailabilityChangeset {
    pub time: (NaiveDateTime, NaiveDateTime),
    pub comment: Option<String>,
}

pub struct AvailabilityContext<'a> {
    pub pool: &'a PgPool,
    pub user_id: i32,
    pub availability_to_get_edited: Option<i32>,
}

impl<'a> AsyncValidate<'a> for AvailabilityChangeset {
    type Context = AvailabilityContext<'a>;

    async fn validate_with_context(
        &self,
        context: &'a Self::Context,
    ) -> Result<(), AsyncValidateError> {
        let mut existing_availabilities =
            Availability::read_by_user_and_date(context.pool, context.user_id, &self.time.0.date())
                .await?;

        if let Some(existing) = context.availability_to_get_edited {
            existing_availabilities = existing_availabilities
                .into_iter()
                .filter(|a| a.id != existing)
                .collect();
        }

        time_is_not_already_made_available(&self.time, &existing_availabilities)?;
        start_date_time_lies_before_end_date_time(&self.time.0, &self.time.1)?;

        Ok(())
    }
}

fn time_is_not_already_made_available(
    value: &(NaiveDateTime, NaiveDateTime),
    existing_availabilities: &Vec<Availability>,
) -> Result<(), AsyncValidateError> {
    if existing_availabilities.is_empty() {
        return Ok(());
    }

    let free_slots = find_free_date_time_slots(existing_availabilities);

    if free_slots.is_empty() {
        return Err(AsyncValidateError::new(
            "cant create a availability as every time slot is already filled",
        ));
    }

    let free_block_found_for_start = free_slots.iter().any(|s| s.0 <= value.0 && s.1 >= value.0);
    let free_block_found_for_end = free_slots.iter().any(|s| s.0 <= value.1 && s.1 >= value.1);

    if !free_block_found_for_start || !free_block_found_for_end {
        return Err(AsyncValidateError::new(
            "cant create availability as there exists already a availability with the desired time",
        ));
    }

    Ok(())
}

pub fn find_free_date_time_slots(
    availabilities: &[Availability],
) -> Vec<(NaiveDateTime, NaiveDateTime)> {
    if availabilities.is_empty() {
        return Vec::new();
    }

    let mut times: Vec<(NaiveDateTime, NaiveDateTime)> =
        availabilities.iter().map(|a| (a.start, a.end)).collect();
    times.sort();

    let mut changed = true;
    while changed {
        changed = false;

        for i in 0..(times.len() - 1) {
            let b = times[i + 1];
            let a = times.get_mut(i).unwrap();

            if a.1 == b.0 {
                a.1 = b.1;
                times.remove(i + 1);
                changed = true;
                break;
            }
        }
    }

    //println!("zeiten unified {times:?}");

    let date = times.first().unwrap().0.date();
    let date_next_day = date.checked_add_days(Days::new(1)).unwrap();
    let start_of_day = date.and_time(START_OF_DAY);
    let end_of_day = date_next_day.and_time(END_OF_DAY);

    // now times contains unified list of existing availabilities -> now calculate the "inverse"

    let mut available_slots = Vec::new();
    let start = times.first().unwrap();
    if start.0 != start_of_day {
        available_slots.push((start_of_day, start.0));
    }

    let mut iterator = times.iter().peekable();
    while let Some(a) = iterator.next() {
        if let Some(b) = iterator.peek() {
            available_slots.push((a.1, b.0));
        }
    }

    let end = times.last().unwrap();
    if end.1 != end_of_day {
        available_slots.push((end.1, end_of_day));
    }

    available_slots
}
