use actix_web::{web, HttpResponse, Responder};
use askama::Template;
use chrono::{NaiveDate, Utc};
use serde::Deserialize;
use sqlx::PgPool;

use crate::{
    filters,
    utils::{
        event_planning_template::generate_vehicles_assigned_and_available,
        ApplicationError,
        DateTimeFormat::{DayMonthYear, DayMonthYearHourMinute, HourMinute},
        TemplateResponse,
    },
};
use brass_db::models::{
    find_free_date_time_slots, Area, Assignment, Availability, Event, Function, Role, User, Vehicle,
};

#[derive(Deserialize)]
pub struct CalendarQuery {
    date: Option<NaiveDate>,
    area: Option<i32>,
}

#[derive(Template)]
#[template(path = "index.html")]
struct CalendarTemplate {
    user: User,
    user_can_create_availability: bool,
    date: NaiveDate,
    selected_area: Option<i32>,
    areas: Vec<Area>,
    events_and_assignments: Vec<(
        Event,
        Vec<String>,
        Option<String>,
        Option<String>,
        Vec<Vehicle>,
    )>,
    availabilities: Vec<Availability>,
}

#[actix_web::get("/")]
async fn get(
    user: web::ReqData<User>,
    pool: web::Data<PgPool>,
    query: web::Query<CalendarQuery>,
) -> Result<impl Responder, ApplicationError> {
    let date = match query.date {
        Some(given_date) => given_date,
        None => Utc::now().date_naive(),
    };

    let areas = Area::read_all(pool.get_ref()).await?;

    let selected_area = match query.area {
        Some(id) => {
            if !areas.iter().any(|a| a.id == id) {
                return Ok(HttpResponse::BadRequest().finish());
            }

            Some(id)
        }
        None => None,
    };

    let availabilities = Availability::read_by_date_and_area_including_user(
        pool.get_ref(),
        date,
        query.area.unwrap_or(user.area_id),
    )
    .await?;

    let availabilities_from_user =
        Availability::read_by_user_and_date(pool.get_ref(), user.id, &date).await?;
    //println!("{availabilities_from_user:#?}");

    let user_can_create_availability = availabilities_from_user.is_empty()
        || !find_free_date_time_slots(&availabilities_from_user).is_empty();

    //println!("{} || {} || {} = {user_can_create_availability}", availabilities_from_user.is_empty(),
    // !only_one_availability_exists_and_is_whole_day(&availabilities_from_user),
    // !find_free_time_slots(&availabilities_from_user).is_empty());

    let mut events_and_assignments = Vec::new();
    for e in Event::read_all_by_date_and_area_including_location(
        pool.get_ref(),
        date,
        query.area.unwrap_or(user.area_id),
    )
    .await?
    .into_iter()
    {
        let assignments = Assignment::read_all_by_event(pool.get_ref(), e.id).await?;
        let (posten, rest): (Vec<Assignment>, Vec<Assignment>) = assignments
            .into_iter()
            .partition(|a| a.function == Function::Posten);
        let (wachhabender, fuehrungsassistent): (Vec<Assignment>, Vec<Assignment>) = rest
            .into_iter()
            .partition(|a| a.function == Function::Wachhabender);

        let (assigned_vehicle, _) = generate_vehicles_assigned_and_available(&pool, &e).await?;

        events_and_assignments.push((
            e,
            posten
                .into_iter()
                .map(|p| {
                    availabilities
                        .iter()
                        .find(|a| a.id == p.availability_id)
                        .unwrap()
                        .user
                        .as_ref()
                        .unwrap()
                        .name
                        .clone()
                })
                .collect(),
            fuehrungsassistent.first().map(|fa| {
                availabilities
                    .iter()
                    .find(|a| a.id == fa.availability_id)
                    .unwrap()
                    .user
                    .as_ref()
                    .unwrap()
                    .name
                    .clone()
            }),
            wachhabender.first().map(|wh| {
                availabilities
                    .iter()
                    .find(|a| a.id == wh.availability_id)
                    .unwrap()
                    .user
                    .as_ref()
                    .unwrap()
                    .name
                    .clone()
            }),
            assigned_vehicle,
        ));
    }

    let template = CalendarTemplate {
        user: user.into_inner(),
        user_can_create_availability,
        date,
        selected_area,
        areas,
        events_and_assignments,
        availabilities,
    };

    Ok(template.to_response()?)
}
