use actix_web::{web, HttpResponse, Responder};
use askama::Template;
use serde_json::json;
use sqlx::PgPool;
use tracing::error;

use crate::{
    endpoints::IdPath,
    filters,
    utils::{ApplicationError, TemplateResponse},
};
use brass_db::models::User;

#[derive(Template)]
#[template(path = "user/profile.html", block = "notificationinput")]
struct ProfilePartialNotificationInputTemplate {
    user: User,
}

#[actix_web::put("/users/{id}/unsubscribeNotifications")]
pub async fn put_unsubscribe(
    user: web::ReqData<User>,
    pool: web::Data<PgPool>,
    path: web::Path<IdPath>,
) -> Result<impl Responder, ApplicationError> {
    handle_subscription_to_notifications(user, pool, path, false).await
}

#[actix_web::put("/users/{id}/subscribeNotifications")]
pub async fn put_subscribe(
    user: web::ReqData<User>,
    pool: web::Data<PgPool>,
    path: web::Path<IdPath>,
) -> Result<impl Responder, ApplicationError> {
    handle_subscription_to_notifications(user, pool, path, true).await
}

async fn handle_subscription_to_notifications(
    user: web::ReqData<User>,
    pool: web::Data<PgPool>,
    path: web::Path<IdPath>,
    subscription_state: bool,
) -> Result<impl Responder, ApplicationError> {
    if user.id != path.id {
        return Err(ApplicationError::Unauthorized);
    }

    let Some(user_in_db) = User::read_by_id(pool.get_ref(), path.id).await? else {
        error!(user = user.id, "Logged in user does not exist in database!");
        return Ok(HttpResponse::NotFound().finish());
    };

    let mut user = user.into_inner();
    if user_in_db.receive_notifications != subscription_state {
        User::update_receive_notifications(pool.get_ref(), user_in_db.id, subscription_state)
            .await?;
        user.receive_notifications = subscription_state;
    }

    let trigger = json!({
        "showToast": {
            "type": "success",
            "message": "Benachrichtigungseinstellung gespeichert!"
        }
    })
    .to_string();

    let template = ProfilePartialNotificationInputTemplate { user };

    Ok(HttpResponse::Ok()
        .insert_header(("HX-TRIGGER", trigger))
        .body(template.to_response()?.into_body()))
}

#[cfg(test)]
mod tests {
    use crate::utils::test_helper::{
        assert_snapshot, read_body, test_put, DbTestContext, RequestConfig, StatusCode,
    };
    use brass_macros::db_test;

    #[db_test]
    async fn user_can_toggle_subscription_for_himself(context: &DbTestContext) {
        let app = context.app().await;

        let unsubscribe_config = RequestConfig::new("/users/1/unsubscribeNotifications");
        let unsubscribe_response =
            test_put::<_, _, String>(&context.db_pool, &app, &unsubscribe_config, None).await;

        assert_eq!(StatusCode::OK, unsubscribe_response.status());

        let unsubscribe_body = read_body(unsubscribe_response).await;
        assert_snapshot!(unsubscribe_body);

        let subscribe_config = RequestConfig::new("/users/1/subscribeNotifications");
        let subscribe_response =
            test_put::<_, _, String>(&context.db_pool, &app, &subscribe_config, None).await;

        assert_eq!(StatusCode::OK, subscribe_response.status());

        let subscribe_body = read_body(subscribe_response).await;
        assert_snapshot!(subscribe_body);
    }

    #[db_test]
    async fn user_cant_toggle_subscription_for_someone_else(context: &DbTestContext) {
        let app = context.app().await;

        let unsubscribe_config = RequestConfig::new("/users/3/unsubscribeNotifications");
        let unsubscribe_response =
            test_put::<_, _, String>(&context.db_pool, &app, &unsubscribe_config, None).await;

        assert_eq!(StatusCode::UNAUTHORIZED, unsubscribe_response.status());
    }
}
