use std::{cell::OnceCell, str::FromStr, sync::Arc, time::Duration};

use actix_http::Request;
use actix_web::{
    body::MessageBody,
    dev::{Service, ServiceResponse},
    test::init_service,
};
use rand::{distr::Alphanumeric, rng, Rng};

use crate::utils::Customization;
use crate::{create_app, mail::Mailer};
use brass_config::{load_config, Config, Environment};
use regex::{Captures, Regex};
use sqlx::{
    postgres::{PgConnectOptions, PgPoolOptions},
    Connection, Executor, PgConnection, PgPool,
};

pub struct DbTestContext {
    pub db_pool: PgPool,
    pub(crate) config: Config,
}

impl DbTestContext {
    pub async fn app(
        &self,
    ) -> impl Service<
        Request,
        Response = ServiceResponse<impl MessageBody>,
        Error = actix_web::error::Error,
    > {
        let customization = Customization {
            webmaster_email: self.config.webmaster_email.clone(),
            hostname: self.config.hostname.clone()
        };

        init_service(create_app(
            self.config.clone(),
            self.db_pool.clone(),
            Mailer::new_stub(),
            customization,
        ))
        .await
    }
}

#[allow(unused)]
pub async fn setup() -> DbTestContext {
    let init_config: OnceCell<Config> = OnceCell::new();
    let config = init_config.get_or_init(|| load_config(&Environment::Test).unwrap());

    let test_db_config = prepare_db(&config.database_url).await;

    let test_db_pool = PgPoolOptions::new()
        .max_lifetime(Some(Duration::from_secs(1))) // workaround for idle connections
        .connect(&test_db_config)
        .await
        .unwrap();

    let context = DbTestContext {
        config: config.clone(),
        db_pool: test_db_pool.clone(),
    };

    context
}

#[allow(unused)]
pub async fn teardown(context: DbTestContext) {
    let mut connect_options = context.db_pool.connect_options();
    let db_config = Arc::make_mut(&mut connect_options);

    drop(context.db_pool);

    let root_db_config = db_config.clone().database("postgres");
    let mut connection: PgConnection = Connection::connect_with(&root_db_config).await.unwrap();

    let test_db_name = db_config.get_database().unwrap();
    // workaround
    //let drop_query = format!("SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE pid <> pg_backend_pid( ) AND datname = {}", test_db_name);
    //connection.execute(drop_query.as_str()).await.unwrap();

    let query = format!("DROP DATABASE IF EXISTS {}", test_db_name);
    //println!("{query}");
    connection.execute(query.as_str()).await.unwrap();
}

pub(crate) async fn prepare_db(url: &str) -> String {
    let db_config = PgConnectOptions::from_str(url).expect("Invalid DATABASE_URL!");
    let db_name = db_config.get_database().unwrap();

    let root_db_config = db_config.clone().database("postgres");
    let mut connection: PgConnection = Connection::connect_with(&root_db_config).await.unwrap();

    let test_db_name = build_test_db_name(db_name);

    let query = format!("CREATE DATABASE {} TEMPLATE {}", test_db_name, db_name);
    connection.execute(query.as_str()).await.unwrap();

    let regex = Regex::new(r"(.+)\/(.+$)").unwrap();
    let test_db_url = regex.replace(url, |caps: &Captures| {
        format!("{}/{}", &caps[1], test_db_name)
    });

    //println!("{test_db_url}");
    test_db_url.to_string()
}

pub(crate) fn build_test_db_name(base_name: &str) -> String {
    let test_db_suffix: String = rng()
        .sample_iter(&Alphanumeric)
        .take(30)
        .map(char::from)
        .collect();
    format!("{}_{}", base_name, test_db_suffix).to_lowercase()
}
